/*	The purpose of this code is, to explore how simple high-pass
 * and low-pass filters can be designed, for use in audio compression,
 * in the limited frequency-range for speech. An 8kHz sample-rate, and
 * 16-bit sample depth are assumed (single channel).
 *
 * 	These filters could be similar to the ones used for
 * 'GSM-EF' or 'AMR' compression, used in cell-phones.
 */


#include <stdlib.h>
#include <stdint.h>

#include "CELP_Filters.h"

#include <stdio.h>


int32_t order2_low_filter(filter_state *state1,
                          filter_state *state2,
                          int32_t in)
{
    int32_t temp_in = 0;
    int32_t temp_out = 0;

    temp_in  = state1->prev_in;
    temp_out = ((state1->k * state1->prev_out) >> 12) +
               (((0x00001000 - state1->k) * ((state1->alpha * in) >> 12)) >> 12);
    state1->prev_in = temp_in;
    state1->prev_out = temp_out;

    temp_in = state2->prev_in;
    temp_out = ((state2->k * state2->prev_out) >> 12) +
               (((0x00001000 - state2->k) * ((state2->alpha * state1->prev_out) >> 12)) >> 12);
    state1->prev_in = temp_in;
    state1->prev_out = temp_out;

    return temp_out;
}


int32_t order_2_high_filter(filter_state *state1,
                            filter_state *state2,
                            int32_t in)
{
    int32_t temp_in;
    int32_t temp_out;

    temp_in  = state1->prev_in;
    temp_out = ((state1->k * state1->prev_out) >> 12) +
               (state1->k * (state1->alpha * (in - temp_in) >> 12) >> 12);
    state1->prev_in = temp_in;
    state1->prev_out = temp_out;

    temp_in = state2->prev_in;
    temp_out = ((state2->k * state2->prev_out) >> 12) +
               (state2->k * (state2->alpha * (state1->prev_out - temp_in) >> 12) >> 12);
    state1->prev_in = temp_in;
    state1->prev_out = temp_out;

    return temp_out;
}

int32_t order_1_high_filter(filter_state *state,
                            int32_t in)
{
    int32_t temp_in;
    int32_t temp_out;

    temp_in  = state->prev_in;
    temp_out = ((state->k * state->prev_out) >> 12) +
               (state->k * (state->alpha * (in - temp_in) >> 12) >> 12);
    state->prev_in = temp_in;
    state->prev_out = temp_out;

    return temp_out;
}


//  First order, high-pass filter, 50Hz
filter_state mike_in = {0x00000F80, 0x00001000, 0, 0};

//  Second order, high-pass filter, 200Hz, State 1
filter_state out_hp1 = {0x00000E00, 0x00001000, 0, 0};

//  Second order, high-pass filter, 200Hz, State 2
filter_state out_hp2 = {0x00000E00, 0x00001000, 0, 0};

//  First LP Corner, Frequency Estimator, 200Hz
filter_state in_fe1 = {0x00000E00, 0x00001000, 0, 0};

//  Second LP Corner, Frequency Estimator, 50Hz
filter_state in_fe2 = {0x00000F80, 0x00001000, 0, 0};

//  First LP Corner, Ouput Integrator, 200Hz
filter_state out_lp1 = {0x00000E00, 0x00004000, 0, 0};

// Second LP Corner, Output Integrator, 1.33kHz
filter_state out_lp2 = {0x00000800, 0x00004000, 0, 0};


/* 	Now, let's test the compiler.
 *
 * 	In the machine language of the i386 and x64 CPUs,
 * Arithmetic Right-Shifts are defined differently from Logical
 * Right-Shifts. At the level of assembler code, an Arithmetic
 * Right-Shift always preserves the value of the MSB while duplicating
 * it into the second-most-significant bit, precisely to preserve the
 * sign of negative integers.
 *
 * 	What this does is open the door to the possibility that a
 * C or C++ compiler can implement the >> operator differently, given
 * 'a >> b' , depending on whether 'a' is a signed integer or an
 * unsigned integer.
 *
 * 	A Logical Right-Shift just commands that the bits be shifted over.
 * By default, an LRS shifts-in a zero bit (unless a certain
 * status-register bit is set on some CPUs).
 *
 * 	Ideal compiler behavior would be, to turn the >> operator
 * into an ARS if 'a' was signed, and, into an LRS if 'a' was unsigned.
 *
 * 	Whether this code will work properly depends, on whether this
 * ideal was realized. Also, certain CPU architectures may exist,
 * which do not define ARS, only LRS. This code performs an ARS, not
 * an LRS, on signed integers, because the operation is faster than
 * integer division.
 *
 * 	AFAIK, ALS and LLS can be treated equally in that, in high-level
 * languages, 'a << b' is Not guaranteed to preserve the sign of 'a'
 * (unless the programmer made sure that the value in 'a' was small
 * enough, to remain a positive value when signed).
 *
 */

int main(int argc, char* argv[])
{
    int16_t s = 0x8000;
    uint16_t u = 0x8000;

    s = s >> 15;
    u = u >> 15;

    printf("The signed integer was right-shifted to become %d\n", s);
    printf("The unsigned integer was right-shifted to become %d\n", u);
    printf("\n");

    if (s == -1 && u == 1) {
        printf("These were the correct results.\n");
    } else {
        printf("Sorry. Your platform doesn't support Arithmetic\n");
        printf("Right-Shift, defined differently from Logical\n");
        printf("Right-Shift.\n");
    }

    printf("\n");
    printf("Type Enter To Quit Program.\n");
    printf("\n");

    getchar();

    return 0;
}
