/*	The purpose of this source file is, to implement 2
 * Phase-Locked Loops, which are to be fed individual signal samples,
 * and which will lock saw-waves 90 degress out of phase with them.
 * 
 * The input samples will be formatted as 32-bit signed integers,
 * that have a 16-bit fractional field.
 * 
 * The internal representation of a state will define the frequency
 * as follows:
 * 
 * 1 - Integer units of frequency, in cycles per 2, 160-sample frames,
 * followed by a 16-bit fractional field, to be compatible with the
 * input sample format.
 * 
 * 2 - Integer accumulator values, that oscillate between
 * -160 and +159 inclusively, with zero fractional bits.
 * 
 * 3 - A divisor, for the product between the saw wave and the
 * input sample, to allow the PLL to remain stable to some
 * lowest frequency.
 * 
 * 
 * The PLLs are to be stable down to frequencies of 8.0 ,
 * which will correspond to 200Hz, if the sample rate is 8kHz.
 * 
 * This also means that signal amplitudes below
 * (10.0 / (1 << 16)) will be ignored, due to the integer
 * arithmetic.
 * 
 * Frequencies greater than 80, which corresponds to 2kHz,
 * will be capped at 80, since a PLL will only work up to
 * 1/2 Nyquist Frequency.
 * 
 * 	This source file cannot be built to run stand-alone .
 * 
 */


#include <stdlib.h>
#include <stdint.h>

#include "CELP_PLL.h"

CELP_PLL_state PLL_state1 = {0, 0, 1600, 3, 0};
CELP_PLL_state PLL_state2 = {0, 0, 1600, 3, 0};
CELP_PLL_state PLL_state3 = {0, 0, 1600, 3, 0};


int32_t CELP_PLL_osc(CELP_PLL_state *state,
					int32_t sample_in)
{
	int temp1 = 0;
	int temp2 = 0;
	int temp3 = 0;
	
	temp1 = state->freq >> 16;
	if (temp1 > 80) {
//		state->freq = 80 << 16;
		state->freq = 0x00500000;
		temp1 = 80;
	}
	
	if (temp1 < 2) {
//		state->freq = 2 << 16;
		state->freq = 0x00020000;
		temp1 = 2;
	}
	
	temp2 = state->accum + temp1;
	temp3 = state->min_freq_mag;
	
	if (temp1 < 1 << temp3)
		temp1 = 1 << temp3;
	
	if (temp2 >= 160) {
		state->freq0 = temp1;  //  Lower Jitter
		temp2 -= 320;
	}
	
	state->accum = temp2;
		
	state->freq += temp2 * sample_in *
					(temp1 >> temp3) / state->divisor;
	
	return (temp2 < 0) ? 0xFFFF0000 : 0x00010000 ;
}

int32_t CELP_PLL_osc_s(CELP_PLL_state *state,
					int32_t sample_in)
{
	int temp1 = 0;
	int temp2 = 0;
	int temp3 = 0;
	
	temp1 = state->freq >> 16;
	if (temp1 > 80) {
//		state->freq = 80 << 16;
		state->freq = 0x00500000;
		temp1 = 80;
	}
	
	if (temp1 < 2) {
//		state->freq = 2 << 16;
		state->freq = 0x00020000;
		temp1 = 2;
	}
	
	temp2 = state->accum + temp1;
	temp3 = state->min_freq_mag;
	
	if (temp1 < 1 << temp3)
		temp1 = 1 << temp3;
	
	if (temp2 >= 160) {
		state->freq0 = temp1;  //  Lower Jitter
		temp2 -= 320;
	}
	
	state->accum = temp2;
		
	state->freq += temp2 * sample_in *
					(temp1 >> temp3) / state->divisor;
	
	return (temp2 < -80) ? 0x00030000 : 0xFFFF0000 ;
}


/*	The suggested way to use the phase-locked loops is as follows:
 * 
 * The standard PLL can be applied twice, to result in a square-wave
 * that is 180 degrees out of phase with the fundamental, because the
 * same 90 degree degree phase shift has been applied twice. The
 * negative of its product with the signal could be called:
 * 
 * amp_sum_f
 * 
 * When a third PLL of the second type is applied to the output of
 * the first PLL, its square wave will be of 1/4 duration and
 * in-phase with the fundamental. Its product with the signal could
 * be called:
 * 
 * amp_sum_st
 * 
 * The sum of the absolutes of the original signal samples
 * was previously called:
 * 
 * input_summation
 * 
 * 
 * 	Then:
 * 
 * tonal_sum = amp_sum_f / 2;
 * 
 * if (tonal_sum < 1) tonal_sum = 1;
 * 
 * if (amp_sum_st < tonal_sum) amp_sum_st = tonal_sum;
 * 
 * length = ((160 * tonal_sum) / amp_sum_st) / frequency;
 * 
 * if (length < 1) length = 1;
 * 
 */
