/*	The purpose of this exercise is:
 *
 * 	1- To Expand a 16-bit selector into a set of up to 4
 * 	sample-psitions, stored in a 5-element array of
 * 	32-bit unsigned integers.
 *
 * 	2- To pack that selector in the first place, given an
 * 	intended maximum length, but, knowing that the way
 * 	the cube root has been fudged, will undershoot the
 * 	intended length when expanded.
 *
 * 	The Selector is going to consist of an 8-bit integer,
 * 	and an 8-bit fractional part.
 *
 * 	If the Selector is to be called (x), then the bit-positions
 * 	1, x, x^2 and x^3  will be marked.
 *
 * 	3- To add an additional number (noise) of bits, that will be
 * 	atonal by virtue of being multiples of a prime number. These
 * 	bits will correspond to the fixed-innovative codeword.
 *
 */

#include <stdint.h>

#include "excite_selector.h"

#include <stdio.h>

#define PRIME 65521U

//  The assumption is being made below, that a pointer
//  is being passed in, to a bit-field that can
//  accomodate bit_pos bits, starting with
//  bit_pos == 0.

//  Little-Endian byte-order is also assumed.


//  Important intial step:
//  Clear garbage from memory...

//  This function will clear an allocated buffer, which is assumed
//  to contain an integer multiple of 32 bits !

int clear_excite_buffer(int buffer_size,
                        uint32_t bit_field[])
{
    int size_chunks = 0;

    if (buffer_size < 1) return -1;

    size_chunks = (buffer_size - 1) >> 5;
    //  An actual value of 0 means 1 chunk...

    for (int i = 0; i <= size_chunks; i++)
        bit_field[i] = 0;

    return 0;
}


int write_excite_buffer(int bit_pos,
                        uint32_t bit_field[])
{
    int element = 0;

    element = bit_pos / 32;
    bit_field[element] |= 1 << (bit_pos % 32);

    return 0;
}

int read_excite_buffer(int bit_pos,
                       uint32_t bit_field[])
{
    int element = 0;
    register int word = 0;

    element = bit_pos / 32;
    word = bit_field[element] >> (bit_pos % 32);

    return (word & 1);
}

//  The assumption is being made below, that a pointer
//  is being passed in, to a bit-field that can
//  accomodate j^3 bits, where j consists of
//  selector's 8 MSBs.

int expand_selector(uint16_t selector,
                    uint32_t bit_field[])
{
    register unsigned int temp = 0x00000100;

    //  Make sure j^3 <= 160.
    if (selector < 0x0100 || selector > 0x0570)
        return -1;

    clear_excite_buffer(160, bit_field);

    for (int i = 0; i < 4; i++) {
        write_excite_buffer((temp >> 8) - 1, bit_field);
        temp *= selector;
        //  Could reach (j^4 << 16) !

        temp = temp >> 8;
    }

    return 0;
}

int pack_selector(int length, uint16_t *out)
{
    register int j = 1;
    int delta = 0;
    int j_cubed = 1;
    register int j1_cubed = 1;

    uint16_t t_out = 0;

    //  To ensure that no constraints will be exceeded,
    //  exit with an error-code, when a codeword longer
    //  than 160 bits has been requested...

    if (length > 160 || length < 1) return -1;

    //  This will also ensure that j <= 5,
    //  and that delta <= 91.
    //  j_cubed <=  125.
    //  j1_cubed <= 216.

    while (j1_cubed <= length) {
        j++;
        j1_cubed = j * j * j;
    }
    j--;
    j_cubed = j * j * j;

    delta = j1_cubed - j_cubed;
    //  delta must not exceed 255 by force of programmer.

    t_out = j << 8;
    //  j must not exceed 255 by force of programmer.

    t_out += ((length - j_cubed) << 8) / delta;
    //  (length - j_cubed) < delta by force of programmer.

    *out = t_out;
    return 0;
}


/* 	The function below will add (noise) bits to the excite buffer.
 * 	The way to compute (noise) is approximately:
 *
 * noise = (resid << 6) / (tonal + resid)
 *
 * ...using integer math.
 *
 * However, since for (noise >= 32), the residual has already
 * drowned out the formant, as soon as (resid >= tonal),
 * what the encoder should do is set:
 *
 * frequency = 1 ,
 * noise = 32
 *
 * ...thereby allowing the full 160-bit excite buffer to be
 * played.
 *
 * 	The bits from the excite codeword should be
 * multiplied by:
 *
 * 	(resid + tonal) * (frequency << 2) /
 * 		((frequency << 2) + noise)
 *
 */


int noise_excite_buffer(int noise, uint32_t bit_field[])
{
    register unsigned int accumulator = 4;  //  Position of first bit.

    for (int i = 0; i < noise; i++) {
        write_excite_buffer(accumulator % 160, bit_field);
        accumulator += (accumulator << 1) + PRIME;
    }

    return 0;
}


/* 	Now, let's test the compiler.
 *
 * 	In the machine language of the i386 and x64 CPUs,
 * Arithmetic Right-Shifts are defined differently from Logical
 * Right-Shifts. At the level of assembler code, an Arithmetic
 * Right-Shift always preserves the value of the MSB while duplicating
 * it into the second-most-significant bit, precisely to preserve the
 * sign of negative integers.
 *
 * 	What this does is open the door to the possibility that a
 * C or C++ compiler can implement the >> operator differently, given
 * 'a >> b' , depending on whether 'a' is a signed integer or an
 * unsigned integer.
 *
 * 	A Logical Right-Shift just commands that the bits be shifted over.
 * By default, an LRS shifts-in a zero bit (unless a certain
 * status-register bit is set on some CPUs).
 *
 * 	Ideal compiler behavior would be, to turn the >> operator
 * into an ARS if 'a' was signed, and, into an LRS if 'a' was unsigned.
 *
 * 	Whether this code will work properly depends, on whether this
 * ideal was realized. Also, certain CPU architectures may exist,
 * which do not define ARS, only LRS. This code performs an ARS, not
 * an LRS, on signed integers, because the operation is faster than
 * integer division.
 *
 * 	AFAIK, ALS and LLS can be treated equally in that, in high-level
 * languages, 'a << b' is Not guaranteed to preserve the sign of 'a'
 * (unless the programmer made sure that the value in 'a' was small
 * enough, to remain a positive value when signed).
 *
 */

int main(int argc, char* argv[])
{
    int16_t s = 0x8000;
    uint16_t u = 0x8000;

    s = s >> 15;
    u = u >> 15;

    printf("The signed integer was right-shifted to become %d\n", s);
    printf("The unsigned integer was right-shifted to become %d\n", u);
    printf("\n");

    if (s == -1 && u == 1) {
        printf("These were the correct results.\n");
    } else {
        printf("Sorry. Your platform doesn't support Arithmetic\n");
        printf("Right-Shift, defined differently from Logical\n");
        printf("Right-Shift.\n");
    }

    printf("\n");
    printf("Type Enter To Quit Program.\n");
    printf("\n");

    getchar();

    return 0;
}
