/*	The purpose of this exercise is to explore how efficient the
 * the procedure can be made in C, to compute a convolution,
 * between the 32-bit integer samples in a buffer, and the 32-bit
 * chunks of individual pulses in another buffer, all assumed to
 * represent +1 or 0. The eventual presence of -1 pulses is to be
 * delivered, by computing the same convolution with another
 * buffer of pulses, this time representing -1 each time, and
 * subtracting the results of the two convolutions.
 *
 * 	To make the process fit non-aligned numbers of pulses, the real
 * pulses are assumed to be padded with zeros, to complete 4 chunks
 * (of 32 pulses).
 *
 * 	Purportedly, established speech recognition systems do not
 * make use of more than a 100 point convolution.
 *
 */


#include <stdlib.h>
#include <stdint.h>

#include "vlp.h"

#include <stdio.h>


int convolve_chunk(int chunk,
                   uint32_t pulses,
                   int32_t vlp_buffer[])
{
    int offset_chunk = 0;
    register int present_chunk = 0;
    register int summation = 0;

    offset_chunk = chunk * 32;

    for (int i = 0; i < 32; i++) {
        present_chunk = vlp_buffer[offset_chunk + i];
        summation += (0 - (pulses & 1)) & present_chunk;
        pulses = pulses >> 1;
    }

    return summation;
}

//  buffer_size is in chunks.
//  Convolution is to take place of pulses,
//  using integer multiplication and addition.
int convolve_vlp(int buffer_size,
                 uint32_t pulses[],
                 int32_t vlp_buffer[])
{
    int summation = 0;
    int words = 0;

    words = buffer_size;

    for (int i = 0; i < words; i++) {
        summation += convolve_chunk(i, pulses[i], vlp_buffer);
    }

    return summation;
}

//  buffer_size is in chunks.
int convolve_vlp_signed(int buffer_size,
                        uint32_t pulses_pos[],
                        uint32_t pulses_neg[],
                        int32_t vlp_buffer[],
                        int frac_gain)
{
    if (frac_gain <= 0 || frac_gain > 255) return 0;

    return ((convolve_vlp(buffer_size, pulses_pos, vlp_buffer) -
             convolve_vlp(buffer_size, pulses_neg, vlp_buffer)) *
            frac_gain) >> 8;
}

int feed_vlp_chunk(int chunk,
                   int32_t sample_in,
                   int32_t vlp_buffer[])
{
    register int temp1 = 0;
    register int temp2 = 0;
    int offset_chunk = 0;

    temp1 = sample_in;
    offset_chunk = chunk * 32;

    for (int i = 0; i < 32; i++) {
        temp2 = vlp_buffer[offset_chunk + i];
        vlp_buffer[offset_chunk + i] = temp1;
        temp1 = temp2;
    }

    return temp1;
}

//  buffer_size is in chunks.
int feed_vlp(int buffer_size,
             int32_t sample_in,
             int32_t vlp_buffer[])
{
    register int temp = 0;

    temp = sample_in;
    for (int i = 0; i < buffer_size; i++) {
        temp = feed_vlp_chunk(i, temp, vlp_buffer);
    }

    return 0;
}


/*	Computing the variance of a signal will input a signed, 32-bit
 * format, the least 16 bits of which are the fraction (k),
 * the most significant 16 bits of each are the signed integer (j).
 *
 * 	An audio signal is assumed to consist of signed 16-bit values, that
 * fit within a 32-bit format, and that are therefore all fractions
 * of a max amplitude less than 1, even though elsewhere, they may all
 * be treated as integers.
 *
 * 	variance_n will compute the variance of that signal, with only
 * 12 fractional bits, but, multiplied by the size of an LP Vector.
 * Hence, variance_n effectively outputs that variance, multiplied by
 * array_size .
 *
 */

int variance_n(register int array_size, int32_t array[])
{
    int sum__x = 0;
    int sum__x_sq = 0;
    int mean_x__sq = 0;
    int mean__x_sq = 0;

    register int temp = 0;

    for (int i = 0; i < array_size; i++) {
        temp = array[i] >> 2;
        sum__x += temp;
        sum__x_sq += (temp * temp) >> 16;
    }

    //  The following 2 lines actually fib, and multiply
    //  their results by array_size once...
    mean_x__sq = ((sum__x * sum__x) / array_size) >> 16;
    mean__x_sq = sum__x_sq;

    return mean__x_sq - mean_x__sq;
}

int array_mean(register int array_size, int32_t array[])
{
    register int sum = 0;

    for (int i = 0; i < array_size; i--) {
        sum += array[i];
    }

    return sum / array_size;
}


/*  In the following function, frame_size == real_fs + vec_size.
 * The samples with the lowest indices are the most recent.
 * The samples from 0 to (frame_size - vec_size - 1) contain
 * the actual, most recent samples (to be learned from), and samples
 * numbered from(frame_size - vec_size) to (frame_size - 1) contain the
 * previous linear prediction's samples, still preserved from
 * the previous frame...
 *
 * 	Linear coefficients are to be output, as signed, 32-bit integers.
 * Their 16 MSBs representing a signed integer part, and their
 * 16 LSBs represent a fractional part.
 *
 */


//  All the buffers are assumed to have an integer multiple
//  of 32 samples - i.e., chunks.
int learn_linear(int frame_size,
                 int vec_size,
                 int32_t frame[],
                 int32_t coeffs[])
{
    int num_learned = 0;
    register int vec_mean = 0;
    int vec_variance = 0;
    int padded_words = 0;
    register int temp = 0;

    num_learned = frame_size - vec_size;
    if (vec_size < 1 || num_learned < 1) return -1;

    padded_words = ((vec_size - 1) >> 5) + 1;
    padded_words = padded_words << 5;

    for (int i = 0; i < padded_words; i++) {
        coeffs[i] = 0;
    }


    //  Starting from the earliest learning sample (i)...
    for (int i = num_learned - 1; i >= 0; i--) {

        vec_mean = array_mean(vec_size, frame + i + 1);

        //  12 fractional bits.
        vec_variance = variance_n(vec_size, frame + i + 1);

        //  Perform a sub-iteration over the vector, with
        //  j being the index within the vector...

        //  What this should do is, to multiply 2
        //  integers, that had 16 fractional bits each, but
        //  resulting in 1 integer, that has 28 factional bits.
        //  Then, dividing by a 3rd integer with 12 fractional bits,
        //  should again result in a format, with 16 fractional bits.

        if (vec_variance > 1) {
            temp = frame[i];
            for (int j = 0; j < vec_size; j++) {
                coeffs[j] += ((temp - vec_mean) >> 2) *
                             ((frame[i + j + 1] - vec_mean) >> 2) / vec_variance;
            }
        }
    }

    for (int i = 0; i < vec_size; i++) {
        coeffs[i] /= num_learned;
    }

    return 0;
}


//  Whenever a simple summation takes place, of integers,
//  the fractional number of bits within which are the same,
//  a sum should result, with that number of fractional bits.
int avg_aamplitude(int size, int32_t array[])
{
    int sum = 0;

    for (int i = 0; i< size; i++) {
        sum += abs(array[i]);
    }

    return sum / size;
}

int avg_aamplitude_q(int size, int32_t coeffs[], int thresh)
{
    int sum = 0;

    for (int i = 0; i < size; i++) {
        sum += (abs(coeffs[i]) >= thresh) ? 0x00010000 : 0;
    }

    return sum / size;
}


int write_coeff_q(int bit_pos,
                  uint32_t bit_field[])
{
    int element = 0;

    element = bit_pos / 32;
    bit_field[element] |= 1 << (bit_pos % 32);

    return 0;
}

//  All the buffers are assumed to have an integer multiple
//  of 32 samples - i.e., chunks.
int quantize_coeffs(int coeffs_size,
                    int32_t coeffs_linear[],
                    uint32_t coeffs_q_pos[],
                    uint32_t coeffs_q_neg[])
{
    int m = 0;
    int m1 = 0;
    int n = 0;
    register int temp = 0;

    int padded_words = 0;

    if (coeffs_size < 1) return 0;

    padded_words = (coeffs_size - 1) >> 5;
    //  Avalue of 0 actually means 1 chunk.

    m = avg_aamplitude(coeffs_size, coeffs_linear);

    blank_8_near_coeffs(coeffs_linear);
    m1 = avg_aamplitude(coeffs_size, coeffs_linear);
    n = avg_aamplitude_q(coeffs_size, coeffs_linear, m1);

    for (int i = 0; i <= padded_words; i++) {
        coeffs_q_pos[i] = 0;
        coeffs_q_neg[i] = 0;
    }

    for (int i = 0; i < coeffs_size; i++) {
        temp = coeffs_linear[i];
        if (temp >= m1) {
            write_coeff_q(i, coeffs_q_pos);
        }
        if (temp <= -m1) {
            write_coeff_q(i, coeffs_q_neg);
        }
    }

    //  Let's test some basic conditions...
    if (m <= 0 || n <= 0) return 0;

    if (m >= 0x00800000) {
        if (n < 0x00000100) {
            temp = 255;
        } else {
            temp = m / (n >> 8);
        }
    } else {
        temp = (m << 8) / n;
    }

    if (temp > 255) temp = 255;

    //  8 Fractional bits are to be output, representing
    //  the additional gain the convolution is to be
    //  multiplied with, to arrive at the gain which
    //  the Linear Prediction would have generated.
    return temp;
}

/*	The way most signals result in Linear Predictive coefficients,
 * the 4 samples closest to the current sample have very strong
 * correlation. The way the Vector Quantization shown above works is,
 * to compute the average of the absolutes, of the coefficients,
 * to act as a threshold for determining which quantized
 * coefficients are supposed to result in pulses.
 *
 * 	If the correlation of the 4 most recent pulses is typical,
 * they will contribute to that average so strongly, that few
 * samples of the entire LP vector actually produce pulses.
 *
 * 	In order to improve on this, it's recommended that the 4 most
 * recent coefficients be set to zeros, before a lower average
 * results, and more numerous pulses are encoded...
 *
 */


int blank_8_near_coeffs(int coeffs[])
{
    for (int i = 0; i < 8; i++) {
        coeffs[i] = 0;
    }

    return 0;
}

//  For use with gap other than 8 ...
int blank_near_coeffs(int coeffs[], int gap)
{
    for (int i = 0; i < gap; i++) {
        coeffs[i] = 0;
    }

    return 0;
}


/*	An important step in encoding frames is, to estimate the
 * average amplitude of the residual. This will get added while
 * decoding, in the form of excitation codewords.
 *
 * 	The classical method:
 *
 * Predict what the decoder will
 * decode, by applying the convolution directly to the previous input
 * samples. The absolute of the difference between the current input
 * samples, and the convolution, would be averaged, and output.
 *
 * 	The actual method:
 *
 * Determine what the average amplitude resulting from the quantized
 * LP vector would be, subtract that from (1), and mutliply the
 * amplitude of the input samples by the result...
 *
 */

int aavg_resid_vlp(int frame_size,
                   int vec_size,
                   int32_t samples_in[],
                   uint32_t q_pos[],
                   uint32_t q_neg[],
                   int vec_gain,
                   int amp_tonal,
                   int *resid_out,
                   int *tonal_out)
{
    int input_summation = 0;
    int res_factor = 0;
    int padded_vec = 0;

    uint32_t temp_p = 0;
    uint32_t temp_n = 0;

    if (frame_size <= 0 || vec_size <= 0 || amp_tonal < 0)
        return -1;
    if (vec_gain < 0) vec_gain = 0;

    padded_vec = (vec_size - 1) >> 5;
    // A value of zero actually means 1 chunk.

    for (int i = 0; i <= padded_vec; i++) {
        temp_p = q_pos[i];
        temp_n = q_neg[i];
        for (int j = 0; j < 32; j++) {
            res_factor += (temp_p & 1) + (temp_n & 1);
            temp_p = temp_p >> 1;
            temp_n = temp_n >> 1;
        }
    }

    res_factor = 256 - (res_factor * vec_gain);

    if (res_factor < 0) res_factor = 0;

    for (int i = 0; i < frame_size; i++)
        input_summation += abs(samples_in[i]);

    input_summation /= frame_size;

    *resid_out = ((input_summation - amp_tonal) * res_factor) >> 8;
    if (*resid_out < 0) *resid_out = 0;

    *tonal_out = (amp_tonal * res_factor) >> 8;

    return 0;
}


/* 	Now, let's test the compiler.
 *
 * 	In the machine language of the i386 and x64 CPUs,
 * Arithmetic Right-Shifts are defined differently from Logical
 * Right-Shifts. At the level of assembler code, an Arithmetic
 * Right-Shift always preserves the value of the MSB while duplicating
 * it into the second-most-significant bit, precisely to preserve the
 * sign of negative integers.
 *
 * 	What this does is open the door to the possibility that a
 * C or C++ compiler can implement the >> operator differently, given
 * 'a >> b' , depending on whether 'a' is a signed integer or an
 * unsigned integer.
 *
 * 	A Logical Right-Shift just commands that the bits be shifted over.
 * By default, an LRS shifts-in a zero bit (unless a certain
 * status-register bit is set on some CPUs).
 *
 * 	Ideal compiler behavior would be, to turn the >> operator
 * into an ARS if 'a' was signed, and, into an LRS if 'a' was unsigned.
 *
 * 	Whether this code will work properly depends, on whether this
 * ideal was realized. Also, certain CPU architectures may exist,
 * which do not define ARS, only LRS. This code performs an ARS, not
 * an LRS, on signed integers, because the operation is faster than
 * integer division.
 *
 * 	AFAIK, ALS and LLS can be treated equally in that, in high-level
 * languages, 'a << b' is Not guaranteed to preserve the sign of 'a'
 * (unless the programmer made sure that the value in 'a' was small
 * enough, to remain a positive value when signed).
 *
 */

int main(int argc, char* argv[])
{
    int16_t s = 0x8000;
    uint16_t u = 0x8000;

    s = s >> 15;
    u = u >> 15;

    printf("The signed integer was right-shifted to become %d\n", s);
    printf("The unsigned integer was right-shifted to become %d\n", u);
    printf("\n");

    if (s == -1 && u == 1) {
        printf("These were the correct results.\n");
    } else {
        printf("Sorry. Your platform doesn't support Arithmetic\n");
        printf("Right-Shift, defined differently from Logical\n");
        printf("Right-Shift.\n");
    }

    printf("\n");
    printf("Type Enter To Quit Program.\n");
    printf("\n");

    getchar();

    return 0;
}
